#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "constype.h"

static char line[LONG_LINE_LENGTH];


static int read_line(FILE *fp, char* line, int *line_no)
{
 if(fgets(line, LONG_LINE_LENGTH, fp)!=NULL)
  {
   (*line_no)++;
   return 1;
  }
 else
   return 0;
}


static void add_signal_by_voltage(SIGNALS *sp, int sig_num, double sig_val, TIME sig_time, int sig_index)
{
 PULSE * pp;


 if(sp->sig_bgn[sig_num] == NULL)
  {
   /* first sample of the first pulse */
   pp=(PULSE*) calloc(1,sizeof(PULSE));
   pp->t = sig_time;
   pp->val = sig_val;
   pp->index = sig_index;
   pp->next = pp->prev = NULL;

   sp->sig_bgn[sig_num]=sp->sig_end[sig_num]=pp;
  }
 else
  {
   if((sig_index - sp->last_index[sig_num]) == 1)
     {
     /* next sample from the same pulse */
      if(sig_val >  sp->sig_end[sig_num]->val)
       {
        /* value closer to the maximum (peak) value */
        sp->sig_end[sig_num]->val = sig_val;
        sp->sig_end[sig_num]->t = sig_time;
        sp->sig_end[sig_num]->index = sig_index;
       }
     }
   else
    /* beginning of next pulse */
     {
      pp=(PULSE*) calloc(1,sizeof(PULSE));
      pp->t=sig_time;
      pp->val=sig_val;
      pp->index= sig_index;
      pp->next=NULL;
   
      pp->prev = sp->sig_end[sig_num];
      sp->sig_end[sig_num]->next = pp;
      sp->sig_end[sig_num] = pp;
     }
  }  

 sp->last_index[sig_num] = sig_index;
}


static void add_signal_by_phase(SIGNALS *sp, int sig_num, double sig_val, TIME sig_time, int sig_index)
{
 PULSE * pp;
 PHASE delta;


 if(sp->sig_bgn[sig_num] == NULL)
  {
   /* first sample of the first pulse */
   pp=(PULSE*) calloc(1,sizeof(PULSE));
   pp->t = pp->t_below = sig_time;
   pp->phase = pp->p_below = sig_val;
   pp->index = sig_index;
   pp->norm_phase= sp->current_phase[sig_num];
   pp->p_above = sp->current_phase[sig_num]+param.max_phase_thresh;
   pp->t_above = 0;
   pp->next = pp->prev = NULL;

   sp->sig_bgn[sig_num]=sp->sig_end[sig_num]=pp;
  }
 else
  {
   if(sp->current_phase[sig_num] == sp->sig_end[sig_num]->norm_phase)
     {
     /* next sample from the same pulse */
      if(sig_val > sp->current_phase[sig_num] + param.max_phase_thresh)
       {
        sp->current_phase[sig_num]+=2*M_PI;
        if(sp->sig_end[sig_num]->t_above==0)
         {
	  if((param.program == INIT_PROGRAM) || 
             ((param.program == HS_PROGRAM) && (param.input_options & M_SET)))
           {
            printf("Warning!\n");
            printf("Only one simulation point has been encountured between PHASE_THRESHOLD_MIN and PHASE_THRESHOLD_MAX!\n");
            printf("The position of the pulse may be estimated with inadequate precision!\n");
            printf("%s: time=%6.2lf\n\n", sp->signal_names[sig_num], sp->sig_end[sig_num]->t_below*1.0e12);
           }
          if(sig_val > sp->sig_end[sig_num]->p_below)
           {
            sp->sig_end[sig_num]->p_above=sig_val; 
            sp->sig_end[sig_num]->t_above=sig_time;
           }
          else
           {
            sp->sig_end[sig_num]->p_above=sp->sig_end[sig_num]->p_below;
            sp->sig_end[sig_num]->t_above=sp->sig_end[sig_num]->t_below;
           }
         }
       }
      else
       {
        if(fabs(sig_val - sp->current_phase[sig_num] - param.nom_phase_thresh) < 
          fabs(sp->sig_end[sig_num]->phase -
               sp->sig_end[sig_num]->norm_phase - param.nom_phase_thresh))
         {
          sp->sig_end[sig_num]->phase=sig_val; 
          sp->sig_end[sig_num]->t=sig_time; 
          sp->sig_end[sig_num]->index=sig_index; 
         }
       if((sig_val <= sp->current_phase[sig_num] + param.nom_phase_thresh) &&
          (sig_val > sp->sig_end[sig_num]->p_below))
         {
          sp->sig_end[sig_num]->p_below=sig_val; 
          sp->sig_end[sig_num]->t_below=sig_time; 
         }
       if((sig_val >= sp->current_phase[sig_num] + param.nom_phase_thresh) &&
          (sig_val < sp->sig_end[sig_num]->p_above))
         {
          sp->sig_end[sig_num]->p_above=sig_val; 
          sp->sig_end[sig_num]->t_above=sig_time; 
         }
       }
     }
   else
    /* beginning of next pulse */
     {
      pp=(PULSE*) calloc(1,sizeof(PULSE));
      pp->t=pp->t_below=sig_time;
      pp->phase=pp->p_below=sig_val;
      pp->index= sig_index;
      pp->norm_phase= sp->current_phase[sig_num];
      pp->p_above = sp->current_phase[sig_num]+param.max_phase_thresh;
      pp->t_above = 0;
      pp->next=NULL;
   
      pp->prev = sp->sig_end[sig_num];
      sp->sig_end[sig_num]->next = pp;
      sp->sig_end[sig_num] = pp;
     }
  }  

}


static int add_edge(SIGNALS *sp, int sig_num, double sig_val, TIME sig_time, double old_sig_val, TIME old_sig_time, int sig_index, int old_sig_index, int direction)
{
 EDGE * pp;
 double delta;

 if(sp->sig_edge_bgn[sig_num] != NULL)
  if(sig_time - sp->sig_edge_end[sig_num]->t < param.min_edge_sep)
   return 0;

   pp=(EDGE*) calloc(1,sizeof(EDGE));
   if(fabs(sig_val - param.nom_edge_thresh) < fabs(param.nom_edge_thresh - old_sig_val)) {
     pp->t = sig_time;
     pp->index = sig_index;
     pp->val = sig_val;
   }
   else {
     pp->t = old_sig_time;
     pp->index = old_sig_index;
     pp->val = old_sig_val;
   }
 pp->direction = direction;
 delta = fabs(old_sig_val - param.nom_edge_thresh)/fabs(sig_val - old_sig_val);
 pp->exact_time = (sig_time-old_sig_time)*delta + old_sig_time;

 if(sp->sig_edge_bgn[sig_num] == NULL) {
   /* first sample of the first pulse */
   pp->next = pp->prev = NULL;
   sp->sig_edge_bgn[sig_num]=sp->sig_edge_end[sig_num]=pp;
 }
 else {
      pp->next=NULL;   
      pp->prev = sp->sig_edge_end[sig_num];
      sp->sig_edge_end[sig_num]->next = pp;
      sp->sig_edge_end[sig_num] = pp;
 }

 return 1;
}


static void remove_not_complete_pulses(SIGNALS *sp)
{
 PULSE *pp;
 int sig_num;

 for(sig_num=0; sig_num<sp->signal_num; sig_num++)
  {
   if(sp->sig_end[sig_num]!=NULL)
    if(sp->sig_end[sig_num]->t_above==0)
     {
/* remove description of the last pulse from the list */
      pp=sp->sig_end[sig_num];
      if(pp->prev==NULL)
       {
        sp->sig_bgn[sig_num]=NULL;
        free(pp);
       }
      else
       {
        pp->prev->next=NULL;
        free(pp);
       }
     }
  }
}

static PHASE normalized_phase(PHASE p_val)
{
 int k;

 k=p_val/(2*M_PI);
 if(p_val<0)
   k--;
 if((p_val - k*(2*M_PI)) < param.nom_phase_thresh)
  return k*(2*M_PI);
 else
  return (k+1)*(2*M_PI);
}


static SIGNALS * collect_pulses(char* filename, int method)
{
 FILE *fp;
 int i,k,m;

 int Sdiv3, Smod3;

 int line_no=0;
 int signals_in_row;
 
 SIGNALS *sp;
 double sig_val;
 double old_sig_val;

 int index, old_index=0;
 float time, old_time;

 int expected_edge[3];


/* open input file for reading */
 fp=fopen(filename, "r");
 if(fp==NULL)
  {
   printf("Cannot open file: %s.\n", filename);
   return NULL;
  }

/* allocate memory for main data structure */
 sp=(SIGNALS*) calloc(1,sizeof(SIGNALS));

/* read number of signals traced */
 read_line(fp, line, &line_no);

 
 if(sscanf(line, "%d %d %d %d", &sp->clk_num, &sp->ins_num, 
                                &sp->inter_num, &sp->outs_num) != 4)
  {
   printf("ERROR IN FILE \".io\"\nset inout_num = %s\n", line);
   return NULL;
  }   
 sp->signal_num=sp->clk_num+sp->ins_num+sp->inter_num+sp->outs_num;

 read_line(fp, line, &line_no);
 i=sscanf(line, "%d", &sp->var_in_nr);
 if(i!=1)
   {
    printf("ERROR IN FILE \".io\"\nset var_in_num = %s\n", line);
    return NULL;
   }   

 sp->sig_initial = (int*) calloc(sp->signal_num, sizeof(int));

/* read the names of signals given by the designer */
 sp->signal_names=(char (*)[NAME_LENGTH])calloc(sp->signal_num,
                                        NAME_LENGTH*sizeof(char));
 read_line(fp, line, &line_no);
 k=0;
 for(m=0; m<sp->signal_num; m++)
   {
    if(sscanf(&line[k], "%s", sp->signal_names[m])!=1)
     {
      printf("ERROR IN FILE \".io\"\nset user_names = %s\n", line);
      return NULL;
     }   
     k+=skip_arg(&line[k], 1);
   }

/* read the names of signals used in jspice simulations */
 if(method==COLLECT_BY_PHASE)
   sp->jspice_phase_names=(char (*)[NAME_LENGTH])calloc(sp->signal_num,
                                        NAME_LENGTH*sizeof(char));
 if((method==COLLECT_BY_VOLTAGE) || (method==COLLECT_EDGES))
   sp->jspice_voltage_names=(char (*)[NAME_LENGTH])calloc(sp->signal_num,
                                        NAME_LENGTH*sizeof(char));

 read_line(fp, line, &line_no);
 k=0;
 for(m=0; m<sp->signal_num; m++) {
    if(method==COLLECT_BY_PHASE) {
      i=sscanf(&line[k], "%s", sp->jspice_phase_names[m]);
      if(i!=1) {
         printf("ERROR IN FILE \".io\"\nset phase_names = %s\n", line);
         return NULL;
      }   
    }

    if((method==COLLECT_BY_VOLTAGE) || (method==COLLECT_EDGES)) {
      i=sscanf(&line[k], "%s", sp->jspice_voltage_names[m]);
      if(i!=1) {
         printf("ERROR IN FILE \".io\"\nset voltage_names = %s\n", line);
         return NULL;
      }   
    }
    
    k+=skip_arg(&line[k], 1);
 }
 

/* allocate memory for necessary arrays */
 sp->sig_bgn=(PULSE**) calloc(sp->signal_num, sizeof(PULSE*));
 sp->sig_end=(PULSE**) calloc(sp->signal_num, sizeof(PULSE*));
 
 if(method==COLLECT_BY_PHASE)
   sp->current_phase=(PHASE*) calloc(sp->signal_num, sizeof(int));

 if(method==COLLECT_BY_VOLTAGE)
   sp->last_index=(int*) calloc(sp->signal_num, sizeof(int));

 if(method==COLLECT_EDGES) {
   sp->sig_edge_bgn=(EDGE**) calloc(sp->signal_num, sizeof(EDGE*));
   sp->sig_edge_end=(EDGE**) calloc(sp->signal_num, sizeof(EDGE*));
 }

/* number of sets of signals (3 signals in the set) */
 Sdiv3=sp->signal_num/3;

/* number of signals in the last set */
 Smod3=sp->signal_num%3;

 if(Smod3!=0)
   Sdiv3++;
 else
   Smod3=3;


 signals_in_row = 3;

 for(i=0; i<=Sdiv3; i++)
  {
   if(i==Sdiv3)
     signals_in_row=Smod3;

   while(read_line(fp, line, &line_no))
    {
     /* if line does not include index and time read next line */
     if(sscanf(line, "%d %e", &index, &time)!=2)
       continue;     

     /* if not subsequent index */
     if(index!=old_index+1)
       if(index==0)
          /* beginning of the next set of signals */
         {
          if(i==0)
           {
            /* store information necessary to compute the time resolution */
            sp->min_time_delta = time;
            sp->start_time = time;
           }

          /* set and check # of valid input samples */
          if(i<2)
           sp->indices_nr=old_index+1;
          else
           if(sp->indices_nr!=old_index+1)
            {
           printf("Warning !!!\n");
           printf("Numbers of samples in different signal sets do not match.");
           printf("Possible error in input data file.");
           sp->indices_nr=MIN(old_index+1, sp->indices_nr); 
            }

          old_index=0;

          goto next_signals;
         }
        else
         continue;

     if((i==1) && (index==1))
       /* compute the time resolution */
       sp->min_time_delta = time - sp->min_time_delta;

     old_index=index;
     k=skip_arg(line, 2);

     /* read values of signals */
     for(m=0; m<signals_in_row; m++)
      {
       if(sscanf(&line[k], "%le", &sig_val)!=1)
         goto next_line;

       switch(method)
        {
         case COLLECT_BY_PHASE:
              if(index==1)
                sp->current_phase[(i-1)*3+m]=normalized_phase(sig_val);

              /* if phase close to the new threshold value
                 check if add new pulse to the list */
              if(sig_val > sp->current_phase[(i-1)*3+m] + 
                      param.min_phase_thresh)
                add_signal_by_phase(sp, (i-1)*3+m, sig_val, time, index); 
              break;

         case COLLECT_BY_VOLTAGE:
              /* if signal value greater than VOLTAGE_THRESHOLD
                 check if add new pulse to the list */
              if(sig_val > param.volt_thresh)
                add_signal_by_voltage(sp, (i-1)*3+m, sig_val, time, index); 
              break;  

         case COLLECT_EDGES:
		if(index==1) {
                  if(sig_val < param.nom_edge_thresh) {
                    sp->sig_initial[(i-1)*3+m] = 0;
                    expected_edge[m] = RISING_EDGE;
                  }
                  else {
                    sp->sig_initial[(i-1)*3+m] = 1;
		    expected_edge[m] = FALLING_EDGE;
                  }
                }

              /* if phase close to the new threshold value
                 check if add new pulse to the list */
	      if(expected_edge[m] == RISING_EDGE) {
                if(sig_val > param.nom_edge_thresh) {
                  if(add_edge(sp, (i-1)*3+m, sig_val, time, old_sig_val, old_time, index, old_index, expected_edge[m]))
		    expected_edge[m] = FALLING_EDGE;
                }
              }
	      else
              if(expected_edge[m] == FALLING_EDGE) {
                if(sig_val < param.nom_edge_thresh) {
                  if(add_edge(sp, (i-1)*3+m, sig_val, time, old_sig_val, old_time, index, old_index, expected_edge[m]))
		    expected_edge[m] = RISING_EDGE;
	        }
              }
	      old_sig_val = sig_val;
	      old_time = time;
              break;
        }


       k+=skip_arg(&line[k], 1);
      }
next_line:  ;
    }
next_signals: ;
  }
 
 if(Sdiv3==1)
   sp->indices_nr=index;

 fclose(fp);

 if(method==COLLECT_BY_PHASE)
   remove_not_complete_pulses(sp);

 return sp;
}



int collect_values(SIGNALS *sp, char* filename, int collected)
{
 FILE *fp;
 int i,k,m, sig_nr;
 PULSE** sig_tmp;

 int Sdiv3, Smod3;

 int line_no=0;
 int signals_in_row, valid_value;
 
 double sig_val;

 int index, old_index=0;
 float time;

 fp=fopen(filename, "r");
 if(fp==NULL)
   return 0;

/* skip number of signals traced */
 read_line(fp, line, &line_no);

/* skip the number of the input with variable generator attached */
 read_line(fp, line, &line_no);

/* skip the names of signals given by the designer */
 read_line(fp, line, &line_no);

/* read the names of signals used in jspice simulations */
 switch(collected)
   {
    case COLLECT_PHASE:          
         sp->jspice_phase_names=(char (*)[NAME_LENGTH])calloc(sp->signal_num,
                                        NAME_LENGTH*sizeof(char));
         break;

    case COLLECT_VOLTAGE:
         sp->jspice_voltage_names=(char (*)[NAME_LENGTH])calloc(sp->signal_num,
                                        NAME_LENGTH*sizeof(char));
         break;
   }

 read_line(fp, line, &line_no);
 k=0;
 for(m=0; m<sp->signal_num; m++)
   {
    if(collected==COLLECT_PHASE)
      i=sscanf(&line[k], "%s", sp->jspice_phase_names[m]);

    if(collected==COLLECT_VOLTAGE)
      i=sscanf(&line[k], "%s", sp->jspice_voltage_names[m]);
     
    if(i!=1)
     {
      printf("ERROR IN LINE: %d\n", line_no);
      return 0;
     }   
     k+=skip_arg(&line[k], 1);
   }


 sig_tmp=(PULSE**) calloc(sp->signal_num, sizeof(PULSE*));
 for(k=0; k<sp->signal_num; k++)
   sig_tmp[k]=sp->sig_bgn[k];

 Sdiv3=sp->signal_num/3;
 Smod3=sp->signal_num%3;
 if(Smod3!=0)
   Sdiv3++;
 else
   Smod3=3;

 signals_in_row = 3;

 for(i=0; i<=Sdiv3; i++)
  {
   if(i==Sdiv3)
     signals_in_row=Smod3;
   while(read_line(fp, line, &line_no))
    {
     if(sscanf(line, "%d %e", &index, &time)!=2)
       continue;     
     if(index!=old_index+1)
       if(index==0)
         {
          old_index=0;
          goto next_signals;
         }
        else
         continue;
     old_index=index;

     valid_value=0;
     for(m=0; m<signals_in_row; m++)
      {
       sig_nr = (i-1)*3+m;
       if(sig_tmp[sig_nr]!=NULL)
        if(time==sig_tmp[sig_nr]->t)
          valid_value=1;
      }
          
     if(!valid_value)
       continue;

     k=skip_arg(line, 2);

     for(m=0; m<signals_in_row; m++)
      {
       if(sscanf(&line[k], "%le", &sig_val)!=1)
         break;

       sig_nr = (i-1)*3+m;
       if(sig_tmp[sig_nr]!=NULL)
        if(time==sig_tmp[sig_nr]->t)
         {
          switch(collected)
           {
            case COLLECT_PHASE:
                 sig_tmp[sig_nr]->phase=sig_val;  
                 break;

            case COLLECT_VOLTAGE:
                 sig_tmp[sig_nr]->val=sig_val;  
                 break;
           }
          sig_tmp[sig_nr]=sig_tmp[sig_nr]->next;
         }
       k+=skip_arg(&line[k], 1);
      }
    }
next_signals: ;
  }

 fclose(fp);
 return 1;
}



static int verify_phase(SIGNALS *sp)
{
 int k;
 PULSE *pp, *last_pp;

 for(k=0; k<sp->signal_num; k++)
  {
   last_pp=sp->sig_bgn[k];
   if(last_pp!=NULL)
     pp=last_pp->next;
   else 
     pp=NULL;

   while(pp!=NULL)
    {
     if(fabs(pp->phase - last_pp->phase)< param.min_pulse_phase_diff)
      /* no 2*PI phase change between peaks
          => smaller pulse removed from the list of pulses */
      if(pp->val > last_pp->val)
       {
        /* remove first pulse */
        if(last_pp->prev != NULL)
          last_pp->prev->next = pp;
        else
          sp->sig_bgn[k]=pp;

        pp->prev = last_pp->prev;
        free(last_pp);

        last_pp=pp;
        pp=pp->next;
       }
      else
       {
        /* remove second pulse */
        if(pp->next!=NULL)
          pp->next->prev = last_pp;
        last_pp->next=pp->next;
        free(pp);
        pp=last_pp->next;
       }
     else
      {      
       last_pp=pp;
       pp=pp->next;
      }
    }
  }
}



int interpolate_phase_thresh_time(SIGNALS *sp_phase)
{
 int k;
 PULSE *pp_phase;
 FILE *fp;

 TIME t10delta;
 PHASE p10delta, pt0delta;


 for(k=0; k<sp_phase->signal_num; k++)
  {
   pp_phase=sp_phase->sig_bgn[k];

   while(pp_phase!=NULL)
    {
     p10delta = pp_phase->p_above-pp_phase->p_below;
     if(p10delta>0)
      {
       pt0delta = pp_phase->norm_phase + param.nom_phase_thresh -
                           pp_phase->p_below;
       t10delta = pp_phase->t_above - pp_phase->t_below;
       pp_phase->t= t10delta*(pt0delta/p10delta) + pp_phase->t_below;
      }
     else
      if(pp_phase->t_above == pp_phase->t_below)
        pp_phase->t=pp_phase->t_above;
      else
        pp_phase->t=(pp_phase->t_above + pp_phase->t_below)/2;
     pp_phase->phase=pp_phase->norm_phase + param.nom_phase_thresh;
     pp_phase=pp_phase->next;
    }
   }

 return 1;
}



SIGNALS *collect_pulses_by_voltage(char *filename)
{
 char phase_filename[NAME_LENGTH];
 SIGNALS *sp;

 sp=collect_pulses(filename, COLLECT_BY_VOLTAGE);
 if(sp!=NULL)
  {
   strcpy(phase_filename, filename);
   strcat(phase_filename, param.phase_ext);
   if(collect_values(sp, phase_filename, COLLECT_PHASE))
     verify_phase(sp);
  }

 return sp;
}


SIGNALS *collect_pulses_by_phase(char *filename)
{
 char phase_filename[NAME_LENGTH];
 SIGNALS *sp;

 strcpy(phase_filename, filename);
 strcat(phase_filename, param.phase_ext);
 
 sp=collect_pulses(phase_filename, COLLECT_BY_PHASE);
 if(sp!=NULL)
  if(param.interpolate)
    interpolate_phase_thresh_time(sp);

 return sp;   
}


SIGNALS *collect_edges(char *filename)
{
 char voltage_filename[NAME_LENGTH];
 SIGNALS *sp;

 strcpy(voltage_filename, filename);
 
 sp=collect_pulses(voltage_filename, COLLECT_EDGES);

#if 0
 if(sp!=NULL)
  if(param.interpolate)
    interpolate_phase_thresh_time(sp);
#endif

 return sp;   
}



void free_data(SIGNALS *sp)
{
 int i;
 PULSE *pp, *ppf;
 ASYNC_CHECK *tmp_check, *next_check;

 if(sp==NULL)
   return;

 for(i=0; i<sp->signal_num; i++)
  {
/* SRW ** error here, these are not atomically allocated.
   if(sp->signal_names!=NULL)
     free(sp->signal_names[i]);
   if(sp->jspice_voltage_names!=NULL)
     free(sp->jspice_voltage_names[i]);
   if(sp->jspice_phase_names!=NULL)
     free(sp->jspice_phase_names[i]);
*/
   if(sp->sig_log!=NULL)
     free(sp->sig_log[i]);

   if(sp->sig_bgn!=NULL)
    {
     pp=sp->sig_bgn[i];
     while(pp!=NULL)
      {
       ppf=pp;
       pp=pp->next;
       free(ppf);
      }
    }
  }
 
 next_check=tmp_check=sp->async_check_list;
 while(tmp_check!=NULL)
  {
   next_check=tmp_check->next;
   free(tmp_check);
   tmp_check=next_check;
  }

 free(sp->signal_names);
 free(sp->jspice_voltage_names);
 free(sp->jspice_phase_names);
 free(sp->sig_log);
 free(sp->sig_bgn);

 free(sp->current_phase);
 free(sp->last_index);
 free(sp->clk_times);
 free(sp->check_index);
}
