#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "constype.h"

static void compute_sync_check_indices(SIGNALS *sp)
{
 PULSE *pp, *pp_next;
 int clk_cycle;

 sp->check_index=(int*)calloc(sp->clk_cycles_nr, sizeof(int));

 pp = pp_next = sp->sig_bgn[0];
 if(pp!= NULL)
   pp_next=pp->next;

 for(clk_cycle=1; (clk_cycle<sp->clk_cycles_nr-1) && (pp_next!=NULL); clk_cycle++)
  {
   sp->check_index[clk_cycle] = (int) (param.checkpoint_ratio * pp_next->index +
                                (1 - param.checkpoint_ratio) * pp->index);
   pp = pp_next;
   pp_next = pp->next;
  }

 sp->check_index[0] = STEP_FORWARD;
 if((pp=sp->sig_bgn[0])!=NULL)
  if((pp_next=sp->sig_bgn[0]->next)!=NULL)
   {
    sp->check_index[0] = 
      pp->index-(1 - param.checkpoint_ratio)*(pp_next->index-pp->index);
    if(sp->check_index[0]<1)
     sp->check_index[0]=MIN(STEP_FORWARD, pp->index);
   }
}


static ASYNC_CHECK* add_async_checkpoint(ASYNC_CHECK* last_checkpoint, 
                                   int index, PHASE phase, int sig_nr)
{
 ASYNC_CHECK* checkpoint;

 checkpoint = last_checkpoint->next = (ASYNC_CHECK*)calloc(1, sizeof(ASYNC_CHECK));
 checkpoint->index = index;
 checkpoint->pi_phase = phase/M_PI;
 if(phase/M_PI - checkpoint->pi_phase > 0.5)
   checkpoint->pi_phase++;
 checkpoint->sig_nr=sig_nr;
 checkpoint->type=REGULAR_CHECKPOINT;
 checkpoint->next=NULL;

 return checkpoint;
}


static void compute_async_check_indices(SIGNALS *sp)
{
 PULSE *pp, *pp_next;
 int clk_cycle, sig_num, out_nr;
 int check1, check2;
 ASYNC_CHECK *last_checkpoint;
 PHASE initial_phase;
 int initial_phase_set;

 
 param.max_index_var = param.max_delay_var/(sp->min_time_delta*1.0e12);

 sp->async_check_index=(ASYNC_CHECK**)calloc(sp->outs_num,
                                             sizeof(ASYNC_CHECK*));

 for(out_nr=0; out_nr<sp->outs_num; out_nr ++)
 {
  last_checkpoint = sp->async_check_index[out_nr] = 
                    (ASYNC_CHECK*)calloc(1, sizeof(ASYNC_CHECK));
  sig_num=sp->signal_num-sp->outs_num+out_nr;

  pp = pp_next = sp->sig_bgn[sig_num];

  if(pp==NULL)
   {
    check1 = (sp->indices_nr-1)-param.max_index_var;
    if(check1 > param.max_index_var)
     {
      sp->initial_phase_index[sig_num] = param.max_index_var;
      last_checkpoint=add_async_checkpoint(last_checkpoint, 
                                           check1, 
                                           0,
                                           sig_num);
     }
    continue;
   }

  pp_next=pp->next;

  initial_phase_set = 0;

  check2 = pp->index - param.max_index_var;
  if(check2 > param.max_index_var)
   {
    initial_phase     = pp->norm_phase;
    initial_phase_set = 1;
    sp->initial_phase_index[sig_num] = param.max_index_var;

    last_checkpoint=add_async_checkpoint(last_checkpoint,
                                           check2,
                                           0,
                                           sig_num);
   }
  else
   {
    printf("Warning!\n");
    printf("First pulse at the output %s is too close to the point where\n",
            sp->signal_names[sig_num]);
    printf("the simulation is started to be traced.\n");
    printf("The phase before this pulse will not be checked.\n");
    printf("position of the pulse = %.2f ps\n\n", pp->t*1.0e12);
   }

  while(pp_next!=NULL)
   {
    if(pp_next->t - pp->t >= 2*param.max_delay_var*1.0e-12)
     {
      check1 = pp->index + param.max_index_var;
      check2 = pp_next->index - param.max_index_var;

      if(!initial_phase_set)
       {
        initial_phase       = pp_next->norm_phase;
        sp->initial_phase_index[sig_num] = check1;
       }

      if(check2>check1)
       {
        if(initial_phase_set)
          last_checkpoint=add_async_checkpoint(last_checkpoint, 
                                               check1, 
                                               pp_next->norm_phase-initial_phase,
                                               sig_num);
        last_checkpoint=add_async_checkpoint(last_checkpoint, 
                                             check2, 
                                             pp_next->norm_phase-initial_phase,
                                             sig_num);
       }
      else
        if(initial_phase_set)       
          last_checkpoint=add_async_checkpoint(last_checkpoint,
                                              (check1+check2)/2,
                                              pp_next->norm_phase-initial_phase,
                                              sig_num);
      initial_phase_set = 1;
     }
    else
      {
       printf("Warning!\n");
       printf("Consecutive output pulses at the output %s\n",
                       sp->signal_names[sig_num]);
       printf(" are too close to define the checkpoints between them.\n");
       printf("position of pulse 1 = %.2f ps\n", pp->t*1.0e12);
       printf("position of pulse 2 = %.2f ps\n\n", pp_next->t*1.0e12);
      }
    pp = pp_next;
    pp_next = pp->next;
   }

  if(!initial_phase_set)
   {
    printf("Warning!!!\n");
    printf("None checkpoint could be set at output %s !!!\n\n",
            sp->signal_names[sig_num]);
    continue;
   }

  if(pp!=NULL)
   {
    check1 = pp->index + param.max_index_var;
    if(check1 < sp->indices_nr)
     {
      last_checkpoint=add_async_checkpoint(last_checkpoint, 
                                           check1, 
                                           pp->norm_phase+2*M_PI-initial_phase,
                                           sig_num);
     }
    else
     {
      printf("Warning!\n");
      printf("Last pulse at the output %s is too close to the point where\n",
              sp->signal_names[sig_num]);
      printf("the simulation is finished to be traced.\n");
      printf("The phase after this pulse will not be checked.\n");
      printf("position of the pulse = %.2f ps\n\n", pp->t*1.0e12);
     }
   }
 }
}


static ASYNC_CHECK* add_to_the_list(SIGNALS *sp, ASYNC_CHECK* last_checkpoint, 
                                    ASYNC_CHECK* check)
{
 ASYNC_CHECK* checkpoint;

 checkpoint =  (ASYNC_CHECK*)calloc(1, sizeof(ASYNC_CHECK));

 checkpoint->index    = check->index;
 checkpoint->pi_phase = check->pi_phase;
 checkpoint->sig_nr   = check->sig_nr;
 checkpoint->type     = check->type;

 if(last_checkpoint!=NULL)
  {
   checkpoint->next      = last_checkpoint->next;
   last_checkpoint->next = checkpoint;
  }
 else
  {
   checkpoint->next     = sp->async_check_list;
   sp->async_check_list = checkpoint;
  }

 return checkpoint;
}


static void make_list_of_async_checks(SIGNALS *sp)
{
 ASYNC_CHECK *first_checkpoint=NULL, *tmp_checkpoint=NULL;
 ASYNC_CHECK **lastcheckpts;
 int all_null;
 int minimum_index, minimum_out_nr;
 int out_nr;


 if(sp->outs_num==1)
  {
   sp->async_check_list=sp->async_check_index[0]->next;
   return;
  }

 lastcheckpts = (ASYNC_CHECK **)calloc(sp->outs_num, sizeof(ASYNC_CHECK *));
 for(out_nr=0; out_nr<sp->outs_num; out_nr++)
   lastcheckpts[out_nr] = sp->async_check_index[out_nr]->next;

 do
  {
   all_null=1;
   minimum_index = sp->indices_nr+1;
   for(out_nr=0; out_nr<sp->outs_num; out_nr++)
    {
     if(lastcheckpts[out_nr]!=NULL)
      {
       if(lastcheckpts[out_nr]->index < minimum_index)
        {
         minimum_index  = lastcheckpts[out_nr]->index;
         minimum_out_nr = out_nr;
        }
       all_null       = 0;
      }
    }

   if(all_null)
     break;

   tmp_checkpoint=add_to_the_list(sp, tmp_checkpoint, lastcheckpts[minimum_out_nr]);


 /*                                  sp->signal_num-sp->outs_num+minimum_out_nr); */
   if(first_checkpoint==NULL)
     first_checkpoint = tmp_checkpoint;
   lastcheckpts[minimum_out_nr]=lastcheckpts[minimum_out_nr]->next;
  }
 while(1);

 sp->async_check_list = first_checkpoint;
}


static int add_to_the_list_of_async_checks(SIGNALS *sp, ASYNC_CHECK* check)
{
 ASYNC_CHECK *tmp_checkpoint;

 tmp_checkpoint = sp->async_check_list;

 if(tmp_checkpoint == NULL)
  {
   add_to_the_list(sp, tmp_checkpoint, check);
   return 1;
  }

 if(check->index < tmp_checkpoint->index)
  {
   add_to_the_list(sp, NULL, check);
   return 1;
  }

 while(tmp_checkpoint->next!=NULL)
  {
   if(tmp_checkpoint->next->index > check->index)
    {
     add_to_the_list(sp, tmp_checkpoint, check);
     return 1;
    }   
   tmp_checkpoint = tmp_checkpoint->next;
  }

 add_to_the_list(sp, tmp_checkpoint, check);
 return 1;
}


static int generate_sync_passfail(SIGNALS *sp, FILE *fp, int runall)
/* returns last check index */
{
 int clk_cycle, sig_num, out_num;
 int *acc_phase;

   /* synchronous circuit */
   compute_sync_check_indices(sp);
   for(sig_num=sp->signal_num-sp->outs_num; sig_num<sp->signal_num;sig_num++)
    {
     fprintf(fp, "* %s:\n", sp->signal_names[sig_num]);
     fprintf(fp, "*  ");
     for(clk_cycle=1; clk_cycle<sp->clk_cycles_nr-1; clk_cycle++)
       fprintf(fp, "%d", sp->sig_log[sig_num][clk_cycle].val);
     fprintf(fp, "\n\n");
    }
   fprintf(fp, "\n\n");


   if(runall) fprintf(fp, "run\n");

   acc_phase = (int*) calloc(sp->outs_num, sizeof(int));


   for(clk_cycle=1; clk_cycle<sp->clk_cycles_nr-1; clk_cycle++)
    {

     if(clk_cycle==1)
      {
       if(!runall) fprintf(fp, "step %d\n",
             sp->check_index[clk_cycle]+STEP_FORWARD);
       fprintf(fp, "\n");
      }
     else
      {
       fprintf(fp, "if chFAIL = 0\n");
       if(!runall) fprintf(fp, "\tstep %d\n",
             sp->check_index[clk_cycle]-sp->check_index[clk_cycle-1]);
      }

     for(sig_num=sp->signal_num-sp->outs_num; sig_num<sp->signal_num;sig_num++)
      {
       out_num = sig_num - (sp->signal_num-sp->outs_num);
       if(sp->sig_log[sig_num][clk_cycle].val > 0)
         acc_phase[out_num]+=2;

       fprintf(fp, "\tif abs(%s[%d] - %s[%d] - %d*pi) > failthres\n",
                   sp->jspice_phase_names[sig_num],
                   sp->check_index[clk_cycle], 
		   sp->jspice_phase_names[sig_num],
		   sp->check_index[0],
                   acc_phase[out_num]);
       fprintf(fp, "\t\tchFAIL = 1\n");
       fprintf(fp, "\t\techo FAILED FOR TIME = %.1f\n", 
                        (sp->min_time_delta*1.0e12)*sp->check_index[clk_cycle]+
                         sp->start_time*1.0e12);
       fprintf(fp, "\tend\n");
      }

     if(clk_cycle!=1)
      fprintf(fp, "end\n\n");
     else
      fprintf(fp, "\n");
    }
   return sp->check_index[sp->clk_cycles_nr-2];
}


static int add_clk_checkpoints(FILE *fp, SIGNALS *sp)
{
 int clk_cycle, sig_num;
 int acc_phase=2;

 for(clk_cycle=1; clk_cycle<sp->clk_cycles_nr-1; clk_cycle++) {
       fprintf(fp, "\tif abs(%s[%d] - %s[%d] - %d*pi) > failthres\n",
                   sp->jspice_phase_names[0],
                   sp->check_index[clk_cycle], 
		   sp->jspice_phase_names[0],
		   sp->check_index[0],
                   acc_phase);
       acc_phase+=2;
       fprintf(fp, "\t\tchFAIL = 1\n");
       fprintf(fp, "\tend\n");
 }
}

static int generate_async_passfail(SIGNALS *sp, FILE *fp, int runall)
/* returns last check index */
{
 int sig_num;
 ASYNC_CHECK *tmp_check;

 int last_index;
 int i, first;

   /* asynchronous circuit */
   compute_async_check_indices(sp);
   make_list_of_async_checks(sp);
   tmp_check = sp->async_check_list;

   fprintf(fp, "* checkpoints:");
   i=0;
   while(tmp_check!=NULL)
    {
     if((i++)%8==0)
       fprintf(fp, "\n*  ");
     fprintf(fp, "%6.1f  ",
               (sp->min_time_delta*1.0e12)*tmp_check->index+
                                           sp->start_time*1.0e12);
     while(tmp_check->next!=NULL)
      {
       if(tmp_check->next->index == tmp_check->index)
         tmp_check=tmp_check->next;
       else
         break;
      }
     tmp_check=tmp_check->next;
    }
   fprintf(fp, "\n\n");

   if(runall) fprintf(fp, "run\n");

   tmp_check = sp->async_check_list;
   first=1;

   while(tmp_check!=NULL)
    {
     if(!first)
      {
       fprintf(fp, "if chFAIL = 0\n");
       if(!runall) fprintf(fp, "\tstep %d\n", tmp_check->index-last_index);
      }
     else
      {
       if(!runall) fprintf(fp, "step %d\n", tmp_check->index+STEP_FORWARD);
      }

     sig_num = tmp_check->sig_nr;
     fprintf(fp, "\tif abs(%s[%d] - %s[%d] - %d*pi) > failthres\n",
                   sp->jspice_phase_names[sig_num],
		   tmp_check->index,
                   sp->jspice_phase_names[sig_num],
		   sp->initial_phase_index[sig_num],
                   tmp_check->pi_phase);
     fprintf(fp, "\t\tchFAIL = 1\n");
     fprintf(fp, "\t\techo FAILED FOR TIME = %.1f\n", sp->start_time*1.0e12 +
                     tmp_check->index*(sp->min_time_delta*1.0e12)); 
     fprintf(fp, "\tend\n");

     while(tmp_check->next!=NULL)
      {
       if(tmp_check->next->index <= tmp_check->index)
        {
         tmp_check=tmp_check->next;
         sig_num = tmp_check->sig_nr;
         fprintf(fp, "\tif abs(%s[%d] - %s[%d] - %d*pi) > failthres\n",
                     sp->jspice_phase_names[sig_num],
		     tmp_check->index,
                     sp->jspice_phase_names[sig_num],
		     sp->initial_phase_index[sig_num],
                     tmp_check->pi_phase);
         fprintf(fp, "\t\tchFAIL = 1\n");
         fprintf(fp, "\t\techo FAILED FOR TIME = %.1f\n", sp->start_time*1.0e12 +
                     tmp_check->index*(sp->min_time_delta*1.0e12)); 
         fprintf(fp, "\tend\n");
        }
       else
        break;
      }

     if(!first)
      fprintf(fp, "end\n\n");
     else
      {
       fprintf(fp, "\n");
       first = 0;
      }

     last_index = tmp_check->index;
     tmp_check=tmp_check->next;
    }

 return last_index;
}



static int add_input_checkpoint(FILE *fp, SIGNALS* sp, int sig_nr, 
                                         int *last_index, int runall)
{
 static input_nr=0;
 PULSE *pp, *pp_prev, *pp_last;
 int pulses_nr=0, check_index;
 int initial_phase_set=0;
 int check_index_set=0;
 ASYNC_CHECK check;


 param.max_index_var = param.max_delay_var/(sp->min_time_delta*1.0e12);

 pp = sp->sig_bgn[sig_nr];

 if(pp==NULL)
  {
   if(sp->indices_nr > 2*param.max_index_var)
    {
     check_index = (sp->indices_nr - 1) - param.max_index_var;
     sp->initial_phase_index[sig_nr] = param.max_index_var;
    }
   else
    {
     printf("Simulation period too short to check for back travelling pulses\n");
     printf("at input %s\n\n", sp->signal_names[sig_nr]);
     return 0;
    }
  }
 else
  {
   while(pp!=NULL)
    {
     pulses_nr++;
     if(!initial_phase_set)
      {
       if(pulses_nr==1)
        {
         if(pp->index > 2*param.max_index_var)
          {
           sp->initial_phase_index[sig_nr]=param.max_index_var;
           initial_phase_set=1;
          }
         else
          {
           printf("Warning!\n");
           printf("First pulse at the input %s is too close to the point where\n",
                  sp->signal_names[sig_nr]);
           printf("the simulation is started to be traced.\n");
           printf("The phase before this pulse will not be checked.\n");
           printf("position of the pulse = %.2f ps\n\n", pp->t*1.0e12);
          }
        }    
       else
        {
         if(pp->index - pp->prev->index > 2*param.max_index_var)
          {
           sp->initial_phase_index[sig_nr]=pp->prev->index+param.max_index_var;
           pulses_nr = 1;
           initial_phase_set=1;
          }
         else
          {
           printf("Warning!\n");
           printf("Consecutive pulses at the input %s\n",
                         sp->signal_names[sig_nr]);
           printf(" are too set the checkpoint between them.\n");
           printf("position of pulse 1 = %.2f ps\n", pp->prev->t*1.0e12);
           printf("position of pulse 2 = %.2f ps\n\n", pp->t*1.0e12);
          }
        }
      }
     if(pp->next==NULL)
       pp_last=pp;
     pp = pp->next;
    }
  }


 if(pulses_nr==0)
  {
   check_index = (sp->indices_nr - 1) - param.max_index_var;
  }
 else
  {   
   pp = pp_last;

   if(sp->indices_nr-pp->index > 2*param.max_index_var)
     check_index = (sp->indices_nr - 1) - param.max_index_var;
   else
    {
     while(pulses_nr > 1)
      {
       pp_prev=pp->prev;

       if(pp->index - pp_prev->index > 2*param.max_index_var)
        {
         check_index = (pp_prev->index + pp->index)/2;
         check_index_set=1;
         pulses_nr--;
         break;
        }
       else
        {
         pp=pp_prev;
         pulses_nr--;
        }
      }
     if(!check_index_set)
      {
       check_index=pp->index-param.max_index_var;
       pulses_nr=0;
      }
    }
  }

 if(check_index > sp->initial_phase_index[sig_nr])
  {
   if(check_index > *last_index)
    {
     if(!runall) fprintf(fp, "\tstep %d\n", check_index - *last_index);
     *last_index = check_index;
    }
   fprintf(fp, "\tif abs(%s[%d] - %s[%d] - %d*pi) > failthres\n",
           sp->jspice_phase_names[sig_nr], 
           check_index,
           sp->jspice_phase_names[sig_nr], 
           sp->initial_phase_index[sig_nr],
           2*pulses_nr);
   fprintf(fp, "\t\tchFAIL = 1\n");
   fprintf(fp, "\tend\n\n");

   check.sig_nr=sig_nr;
   check.type = REGULAR_CHECKPOINT;
   check.index = check_index;
   check.pi_phase = 2*pulses_nr;
   add_to_the_list_of_async_checks(sp, &check);
  }
 else
  {
   printf("Warning!!!\n");
   printf("Test for back travelling pulses on input %s cannot be performed because the pulses are too close to each other.\n\n", sp->signal_names[sig_nr]);
   return 0;
  }

 return 1;
}


int generate_passfail_file(SIGNALS *sp, char *filename, int runall)
{
 FILE *fp;
 int sig_num;
 int last_index;
 int i, first;
 ASYNC_CHECK *tmp_check;

 fp=fopen(filename, "w");
 if(fp==NULL)
   return 0;

 fprintf(fp, "%s\n", filename);
 fprintf(fp, ".control\n\n");

 sp->initial_phase_index=(int*)calloc(sp->signal_num, sizeof(int));

 if(sp->clk_num>0)
   last_index=generate_sync_passfail(sp, fp, runall);
 else
   last_index=generate_async_passfail(sp, fp, runall);

 fprintf(fp, "\nif chFAIL = 0\n");
 fprintf(fp, "if check_inputs = 1\n");

 if(param.program != HS_PROGRAM) {
 if(sp->clk_num>0) {
   add_input_checkpoint(fp, sp, 0, &last_index, runall);
   add_clk_checkpoints(fp, sp);
 }
 for(sig_num=1; sig_num<sp->clk_num+sp->ins_num; sig_num++)
    add_input_checkpoint(fp, sp, sig_num, &last_index, runall);
 }
 else {
  if(sp->clk_num>0)
   add_clk_checkpoints(fp, sp);
 }

 fprintf(fp, "\tif chFAIL = 1\n");  
 fprintf(fp, "\t\techo FAILED BECAUSE OF EXTRA INPUT PULSES\n");
 fprintf(fp, "\tend\n");
 fprintf(fp, "end\n");
 fprintf(fp, "end\n\n");

 fprintf(fp, "if chFAIL = 1\n");
 fprintf(fp, "\tfree all yes\n");
 fprintf(fp, "\tchFAIL = 1\n");
 fprintf(fp, "end\n\n");

 fprintf(fp, "if chFAIL = 0\n");
 fprintf(fp, "\tfree all yes\n");
 fprintf(fp, "\tchFAIL = 0\n");
 fprintf(fp, "end\n\n");

 return 1;
}


static void add_initial_checkpoints(SIGNALS *sp)
{
 ASYNC_CHECK check;
 int sig_num, last_sig;

 for(sig_num=0; sig_num<sp->ins_num+sp->clk_num; sig_num++)
  {
   if(sp->initial_phase_index[sig_num]>0)
    {
     check.sig_nr=sig_num;
     check.type = INIT_CHECKPOINT;
     check.index = sp->initial_phase_index[sig_num];
     check.pi_phase = 0;
     add_to_the_list_of_async_checks(sp, &check);
    }
  }

 if(sp->clk_num==0)
  {
   for(sig_num=sp->signal_num-sp->outs_num; sig_num<sp->signal_num; sig_num++)
    {
     if(sp->initial_phase_index[sig_num]>0)
      {
       check.sig_nr=sig_num;
       check.type = INIT_CHECKPOINT;
       check.index = sp->initial_phase_index[sig_num];
       check.pi_phase = 0;
       add_to_the_list_of_async_checks(sp, &check);
      }
    }
  }
 else
  {
   for(sig_num=sp->signal_num-sp->outs_num; sig_num<sp->signal_num; sig_num++)
    {
     check.sig_nr=sig_num;
     check.type = INIT_CHECKPOINT;
     check.index = sp->check_index[0];
     check.pi_phase = 0;
     add_to_the_list_of_async_checks(sp, &check);
    }
  }
}


static void add_sync_checkpoints(SIGNALS *sp)
{

 int clk_cycle, sig_num, out_num, check_nr;
 int *acc_phase;
 ASYNC_CHECK check;

 acc_phase = (int*) calloc(sp->outs_num, sizeof(int));

 for(clk_cycle=1; clk_cycle<sp->clk_cycles_nr-1; clk_cycle++)
  {
   for(sig_num=sp->signal_num-sp->outs_num; sig_num<sp->signal_num;sig_num++)
    {
     out_num = sig_num - (sp->signal_num-sp->outs_num);

     if(sp->sig_log[sig_num][clk_cycle].val > 0)
       acc_phase[out_num]+=2;

     check.sig_nr=sig_num;
     check.type = REGULAR_CHECKPOINT;
     check.index = sp->check_index[clk_cycle];
     check.pi_phase = acc_phase[out_num];
     add_to_the_list_of_async_checks(sp, &check);
    }
  }
}


int generate_check_file(SIGNALS *sp, char *filename)
{
 FILE *fp;
 int i;
 ASYNC_CHECK *tmp_check;
 int sig_num, out_num, check_nr;


 fp=fopen(filename, "w");
 if(fp==NULL)
   return 0;

 add_initial_checkpoints(sp);
 if(sp->clk_num>0)
   add_sync_checkpoints(sp);


 fprintf(fp, ".check\n\n");

 for(i=1; i<=2; i++)
  {
   fprintf(fp, "checkVAL%d = \n", i);
   fprintf(fp, "checkSTP%d = \n", i);
   fprintf(fp, "checkDEL%d = \n\n", i);
  }

 fprintf(fp, "compose checkPNTS values ");

 tmp_check = sp->async_check_list;

 while(tmp_check!=NULL)
   {
    fprintf(fp, "%6.1fps  ",
               (sp->min_time_delta*1.0e12)*tmp_check->index+
                                           sp->start_time*1.0e12);
    while(tmp_check->next!=NULL)
     {
      if(tmp_check->next->index == tmp_check->index)
         tmp_check=tmp_check->next;
      else
         break;
     }
     tmp_check=tmp_check->next;
   }
 fprintf(fp, "%6.1fps\n\n",
             (sp->min_time_delta*1.0e12)*(sp->indices_nr-1)+
                                          sp->start_time*1.0e12);


 fprintf(fp, "checkFAIL = 0\n");
 
 fprintf(fp, "failthres = %4.2f\n", param.min_fail_thresh);

 fprintf(fp, "source %s%s%s\n\n", param.input_dir,
                                  param.circuit_name,
                                  param.param_ext);

 fprintf(fp, ".control\n\n");

 tmp_check = sp->async_check_list;
 check_nr  = 0;

 while(tmp_check!=NULL)
  {
    fprintf(fp, "if TIME > checkPNTS[%d]\n", check_nr);
    fprintf(fp, "\tif TIME < checkPNTS[%d]\n", check_nr+1);

    sig_num = tmp_check->sig_nr;
    if(tmp_check->type==REGULAR_CHECKPOINT)
     {
      fprintf(fp, "\t\tif abs(%s - phase0[%d] - %d*pi) > failthres\n",
                   sp->jspice_phase_names[sig_num],
		   sig_num,
                   tmp_check->pi_phase);
      fprintf(fp, "\t\t\tcheckFAIL = 1\n");
      fprintf(fp, "\t\tend\n");
     }
    if(tmp_check->type==INIT_CHECKPOINT)
     {
      fprintf(fp, "\t\tphase0[%d] = %s\n",
               tmp_check->sig_nr,
               sp->jspice_phase_names[tmp_check->sig_nr]);
     }

    while(tmp_check->next!=NULL)
     {
       if(tmp_check->next->index == tmp_check->index)
        {
         tmp_check=tmp_check->next;
         sig_num = tmp_check->sig_nr;
         if(tmp_check->type==REGULAR_CHECKPOINT)
          {
           fprintf(fp, "\t\tif abs(%s - phase0[%d] - %d*pi) > failthres\n",
                   sp->jspice_phase_names[sig_num],
		   sig_num,
                   tmp_check->pi_phase);     
           fprintf(fp, "\t\t\tcheckFAIL = 1\n");
           fprintf(fp, "\t\tend\n");    
          }

         if(tmp_check->type==INIT_CHECKPOINT)
          {
           fprintf(fp, "\t\tphase0[%d] = %s\n",
                   tmp_check->sig_nr,
                   sp->jspice_phase_names[tmp_check->sig_nr]);
          }
        }
       else
        break;
     }

    fprintf(fp, "\tend\n");
    fprintf(fp, "end\n\n");

    tmp_check=tmp_check->next;
    check_nr++;
  }
 

 fprintf(fp, ".endc\n");
 fprintf(fp, "%s\n\n", param.circuit_name);
 fprintf(fp, ".include %s%s%s%s\n\n", param.input_dir,
                                      param.circuit_name,
                                      param.circuit_ext,
                                      param.check_ext);
 fprintf(fp, ".end\n");

 return 1;
}

