#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "constype.h"
#include "filenames.h"

#include "qnrutil.h"

int dim, var = 0, *yield, pin = 0, *xyz;
int pntcount = 0;

double  *startpnt, *centerpnt, *lower, *upper, *scale, *delta, **hullpnts;
char (*names)[NAME_LENGTH];

PARAMETER_SET *parameter_list;


static int skip_arg(char * line, int arg_num)
{
 int i=0, k;
 int line_length = strlen(line);

 for(k=0; k<arg_num; k++)
  {
   while(isspace(line[i]) && (i<line_length))
     i++;
   while(!isspace(line[i]) && (i<line_length))
     i++;
  }

 while(isspace(line[i]) && (i<line_length))
     i++;

 return i;
}


int addpoint(double *pc,double *po)
{
        FILE *fp;
	double dum, dist=0.0, inveffcenter;
	int x, concave;
	char errmsg[LINE_LENGTH];
        char system_call[LINE_LENGTH];

        if ((fp=fopen(LIMITS_FILE,"w")) == NULL) {
                strcpy(errmsg, "Can't write to the '");
		strcat(errmsg, LIMITS_FILE);
		strcat(errmsg, "' file.\n");
                nrerror(errmsg);
        }
/* load zeroeth element with unscaled value for accuracy */
        for (x=1; dim >= x; ++x)
                dist += (pc[x]-po[x])*(pc[x]-po[x])*scale[x]*scale[x];
	dist = sqrt(dist);

        inveffcenter = fabs(pc[1]-po[1])/(dist*centerpnt[1]);
        for (x=2; dim >= x; ++x)
                if((dum = fabs(pc[x]-po[x])/(dist*centerpnt[x])) > inveffcenter)
			 inveffcenter = dum;
/* in 1-D, X=(pc[]-po[])/centerpnt[ ] */
/* in 2-D, pick effcenter such that above holds for each dimension */
        fprintf(fp,"pc[0]=0\npo[0]=%f\n",dist*inveffcenter);
/* pc on plane, po on the boundary */
/* unscale it */
	for (x=1; dim >= x; ++x) {
		fprintf(fp, "pc[%i]=%f\n", x, pc[x]*scale[x]);
		fprintf(fp, "po[%i]=%f\n", x, po[x]*scale[x]);
	}
/* print the circuit and circnumber names */
	fprintf(fp, "set circuit = '%s'\n", param.circuit_name);
	fprintf(fp, "set opt_num = '%s'\n", param.opt_num);
	fclose(fp);

        strcpy(system_call, param.spice_name);
        strcat(system_call, BOUNDARY_CALL);
	system(system_call);

        if ((fp=fopen(POINT_FILE,"r")) == NULL) {
                strcpy(errmsg, "Can't read the '");
		strcat(errmsg, POINT_FILE);
		strcat(errmsg, "' file.\n");
                nrerror(errmsg);
        }
	fscanf(fp, "%d", &concave);

/* read in and scale the new point */
	if (!concave) {
		++pntcount;

                /* throw away the zeroeth array element */
	        fscanf(fp, "%lf",&dum);
		for (x=1; dim >= x; ++x) {
			fscanf(fp, "%lf", *(hullpnts+pntcount)+x);
			hullpnts[pntcount][x] /= scale[x];
                }

		for (x=0; dim >= x; ++x) 
			fscanf(fp, "%lf", &delta[x]);
	}

	fclose(fp);
	return(!concave);
}

static int read_long_line(FILE *fp, char* line, int *line_no)
{
 if(fgets(line, LONG_LINE_LENGTH, fp)!=NULL)
  {
   (*line_no)++;
   return 1;
  }
 else
   return 0;
}


static void error_in_line(char* filename, int line_num, char*line)
{
 printf("ERROR IN FILE %s LINE: %d\n%s\n", filename, line_num, line);
 exit(255);
}


static void unexpctd_eof(char* filename)
{
 printf("UNEXPECTED END OF FILE: %s\n", filename);
 exit(255);
}


static int blank_line(char *line)
{
 int i=0, n=strlen(line);

 while(i<n)
  {
   if(!isspace(line[i++]))
     return 0;
  }
 return 1;
}

static void add_parameter_set(PARAMETER_SET *new)
{
 PARAMETER_SET *tmp;
 static PARAMETER_SET *end_of_list;

 if(parameter_list==NULL)
  end_of_list=parameter_list=new;
 else {
  end_of_list->next=new;
  end_of_list=new;
 }
}

void readinit(void)
{
        FILE *fp;
	char filename[LINE_LENGTH];
        int x,k,temp;
	char in_buffer[LONG_LINE_LENGTH],errmssg[LINE_LENGTH+30];
	int line_num=0;

/* string out the input file name */
	strcpy(filename, param.input_dir);
	strcat(filename, param.circuit_name);
	strcat(filename, param.init_ext);
	strcat(filename, param.opt_num);

/* string out the error message */
	strcpy(errmssg, "Can't read the '");
	strcat(errmssg, filename);
	strcat(errmssg, "'file.\n");

/* read init file */
        if ((fp=fopen(filename, "r")) == NULL)
                nrerror(errmssg);

	do {
	    if(!read_long_line(fp, in_buffer, &line_num))
              unexpctd_eof(filename);
	} while(in_buffer[0]=='*');
	if(sscanf(in_buffer, "%d", &dim)!=1)
		error_in_line(filename, line_num, in_buffer);

/* initialize */
	startpnt  = vector(1,dim);
	centerpnt = vector(1,dim);
        lower	  = vector(1,dim);
        upper     = vector(1,dim);
        scale     = vector(1,dim);
        yield     = ivector(1,dim);
	names   = (char(*)[NAME_LENGTH])calloc(dim, NAME_LENGTH);
        delta = (double *)calloc(dim+1, sizeof(double));

/* read some more */

    /* parameter names */
	do {    
	    if(!read_long_line(fp, in_buffer, &line_num))
              unexpctd_eof(filename);
	} while((in_buffer[0]=='*') || blank_line(in_buffer));
	k=0;
	for (x=0; dim > x; ++x) {
		if(sscanf(&in_buffer[k], "%s", names[x])!=1)
			error_in_line(filename, line_num, in_buffer);
		k+=skip_arg(&in_buffer[k], 1);
	}	

    /* nominal values */
	do { 
	    if(!read_long_line(fp, in_buffer, &line_num))
              unexpctd_eof(filename);
	} while((in_buffer[0]=='*') || blank_line(in_buffer));
	k=0;
	for (x=1; dim >= x; ++x) {
		if(sscanf(&in_buffer[k], "%lf", centerpnt+x)!=1)
			error_in_line(filename, line_num, in_buffer);
		k+=skip_arg(&in_buffer[k], 1);
	}

    /* minimum parameter search values */
	do {   
	    if(!read_long_line(fp, in_buffer, &line_num))
              unexpctd_eof(filename);
        } while((in_buffer[0]=='*') || blank_line(in_buffer));
	k=0;
	for (x=1; dim >= x; ++x) {
		if(sscanf(&in_buffer[k], "%lf", lower+x)!=1)
			error_in_line(filename, line_num, in_buffer);
		k+=skip_arg(&in_buffer[k], 1);
	}

    /* maximum parameter search values */
        do {   
	    if(!read_long_line(fp, in_buffer, &line_num))
              unexpctd_eof(filename);
        } while((in_buffer[0]=='*') || blank_line(in_buffer));
	k=0;
	for (x=1; dim >= x; ++x) {
		if(sscanf(&in_buffer[k], "%lf", upper+x)!=1)
			error_in_line(filename, line_num, in_buffer);
		k+=skip_arg(&in_buffer[k], 1);
	}

	if(param.program == OPT_PROGRAM) {
     /* percent variations */
	 do {
            if(!read_long_line(fp, in_buffer, &line_num))
              unexpctd_eof(filename);
         } while((in_buffer[0]=='*') || blank_line(in_buffer));
         k=0;
         for (x=1; dim >= x; ++x) {
                if(sscanf(&in_buffer[k], "%lf", scale+x)!=1)
                        error_in_line(filename, line_num, in_buffer);
                k+=skip_arg(&in_buffer[k], 1);
         }

     /* yield */
         do {
            if(!read_long_line(fp, in_buffer, &line_num))
              unexpctd_eof(filename);
         } while((in_buffer[0]=='*') || blank_line(in_buffer));
         k=0;
         for (x=1; dim >= x; ++x) {
                if(sscanf(&in_buffer[k], "%d", yield+x)!=1)
                        error_in_line(filename, line_num, in_buffer);
                k+=skip_arg(&in_buffer[k], 1);

		if (yield[x])  ++var;
         }

     /* static */
         do {
             if(!read_long_line(fp, in_buffer, &line_num))
               unexpctd_eof(filename);
         } while(in_buffer[0]=='*');
         k=0;
         for (x=1; dim >= x; ++x) {
                 if(sscanf(&in_buffer[k], "%d", &temp)!=1)
                         error_in_line(filename, line_num, in_buffer);
                 k+=skip_arg(&in_buffer[k], 1);
 
                 if (temp)  ++pin;
         }
         xyz = ivector(1,pin);
         pin=0;
         k=0;
         for (x=1; dim >= x; ++x) {
                 if(sscanf(&in_buffer[k], "%d", &temp)!=1)
                         error_in_line(filename, line_num, in_buffer);
                 k+=skip_arg(&in_buffer[k], 1);
 
                 if (temp)
                         xyz[++pin]=x;
          }
	 } /* endif OPT_PROGRAM */

    
     /* 2D - mapping (optional) */
	if((param.program == MARG_PROGRAM) && (param.input_options & TWO_SET)) {
	 PARAMETER_SET *tmp;

	 while(read_long_line(fp, in_buffer, &line_num)) {
           if((in_buffer[0]=='*') || blank_line(in_buffer))
             continue;
	   tmp=(PARAMETER_SET*)calloc(1, sizeof(PARAMETER_SET));
	   if(sscanf(in_buffer, "%d %d %d", &tmp->x, 
					    &tmp->y,
 					    &tmp->points_nr)!=3) {
		free(tmp);
	     	error_in_line(filename, line_num, in_buffer);
	   }
	   else {
	    if((tmp->points_nr<4) || (tmp->x<1) || (tmp->y<1) ||
		(tmp->x>dim) || (tmp->y>dim) || (tmp->x==tmp->y))
		error_in_line(filename, line_num, in_buffer);
	   }
	   add_parameter_set(tmp);
	 }
	} /* endif TWO_SET */

	if((param.program == MARG_PROGRAM)||(param.program == HS_PROGRAM))
        	for (x=1; dim >= x; ++x)
                	scale[x] = 1.0;
/* scale it */
        for (x=1; dim >= x; ++x) {
                scale[x] *= fabs(centerpnt[x]);
                centerpnt[x] /= scale[x];
                lower[x] /= scale[x];
                upper[x] /= scale[x];
                startpnt[x] = centerpnt[x];
        }

	fclose(fp);
}
