#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "constype.h"

static void print_condition(FILE *fp, char * name, TIME chtime, int index, int lvalue, int printstep, int stepnr)
{
 char sign = lvalue?'<':'>';

 fprintf(fp, "if chFAIL = 0\n");
 if(printstep)
   fprintf(fp, "\tstep %d\n", stepnr);
 fprintf(fp, "\tif (%s[%d]) %c vthreshold%d\n", name, index, sign, lvalue);
 fprintf(fp, "\t\tchFAIL = 1\n");
 fprintf(fp, "\t\techo FAILED FOR TIME = %.1f ps\n", chtime*1.0e12);
 fprintf(fp, "\tend\n");
 fprintf(fp, "end\n\n");
}


static int latch_signal_value(TIME check_time, EDGE **prev_edge, int init_value)
{
 EDGE *ne, *pe;

 pe=(*prev_edge);
 if(pe == NULL)
   return init_value;

 if(pe->exact_time > check_time)
   return init_value;

 ne=pe->next;
 while(ne!=NULL) {
   if(check_time < ne->exact_time)
     break;
/* if the difference is too small - WARNING !!! */

  pe=ne;
  ne=ne->next;
 }

 (*prev_edge)=pe;
 return 1 - pe->direction;
}


static int generate_sync_latch_passfail(FILE *fp, SIGNALS *sp, int runall)
{
 EDGE **prev_edge, *clk_edge;
 TIME check_time, check_norm_time, clk_pulse_bgn, clk_pulse_end, total_time;
 int check_index;
 int signal_value;
 int sig_num;
 int step_nr;
 int last_index=-STEP_FORWARD;
 int printstep;

 total_time = sp->start_time+sp->indices_nr*sp->min_time_delta;

/* preparing an array of output signal edges prior to the clock checkpoint */
 prev_edge = (EDGE **) calloc(sizeof(EDGE *), sp->outs_num);
 if(prev_edge == NULL) {
  exit(255); /* WARNING!!! */
 }
 for(sig_num=0; sig_num<sp->outs_num; sig_num++) {
  prev_edge[sig_num] = sp->sig_edge_bgn[sig_num+sp->signal_num-sp->outs_num];
 }

/* starting from the first edge of the clock signal */
 clk_edge = sp->sig_edge_bgn[0];

 do {
   if(clk_edge==NULL)
     break;
    
   if(clk_edge->direction == RISING_EDGE) {
     clk_pulse_bgn = clk_edge->exact_time;
     clk_edge=clk_edge->next;
     if(clk_edge!=NULL)
       clk_pulse_end = clk_edge->exact_time;
     else
       clk_pulse_end = total_time;
    }
   else
/* if the first edge is a falling edge, place the checkpoint between the 
   start of the simulation and the first edge */
   if(clk_edge->direction == FALLING_EDGE) {
     clk_pulse_bgn = sp->start_time;
     clk_pulse_end = clk_edge->exact_time;
    }

   check_time = clk_pulse_bgn + 
               (clk_pulse_end - clk_pulse_bgn)*param.clk_chktime_ratio;
   check_index = (check_time-sp->start_time)/sp->min_time_delta;
   check_norm_time = sp->start_time + check_index*sp->min_time_delta;

   printstep = !runall;
   for(sig_num=sp->signal_num-sp->outs_num; sig_num<sp->signal_num; sig_num++) {
     signal_value = 
       latch_signal_value(check_time, 
                          &prev_edge[sig_num-sp->signal_num+sp->outs_num],
                          sp->sig_initial[sig_num]);
       print_condition(fp, sp->jspice_voltage_names[sig_num], check_norm_time, check_index, signal_value, printstep, check_index-last_index);
    printstep = 0; 
    /* steps executed only before the first output signal is checked */
   }

   last_index = check_index;
    
   if(clk_edge!=NULL)
      clk_edge=clk_edge->next;
   } while (1);
}


#define STEP_PARTS 10

struct passfail_condition {
 int index;  /* index corresponding to the checkpoint */
 int signum; /* signal number */
 TIME time;  /* checkpoint time */
 int value;  /* logical value of the signal */
 struct passfail_condition * next;  /* pointer to the next component of the list */
};

typedef struct passfail_condition PF_CONDITION;

static int collect_passfail_condition(PF_CONDITION ** bpf, int sig_num, TIME check_time, int curr_index, int curr_value)
/* preparing a list of passfail conditions */
{
 PF_CONDITION *npf, *cpf, *opf;

 npf=(PF_CONDITION *) calloc(1, sizeof(PF_CONDITION));
 if(npf==0)
   return 0;
 npf->signum = sig_num;
 npf->index = curr_index;
 npf->time = check_time;
 npf->value = curr_value;

/* empty list */
 if((*bpf) == NULL) {
  (*bpf) = npf;
  npf->next = NULL;
  return 1;
 }
 
 opf=cpf=(*bpf);

/* new component introduced at the beginning of the list */
 if(curr_index < cpf->index) {
  (*bpf) = npf;
  npf->next = cpf;
  return 1;
 }

/* looking for an appropriate position on the list */
 while(cpf!=NULL) {
  if(curr_index < cpf->index)
    break;
  opf=cpf;
  cpf=cpf->next;
 }
 opf->next = npf;
 npf->next = cpf;
 return 1;
}


static void print_passfail_conditions(FILE *fp, SIGNALS*sp, PF_CONDITION *bpf, int runall)
/* printing a list of passfail conditions */
{
 PF_CONDITION *cpf=bpf;
 int i;
 int stepnum, indexlimit, edgenum;

 if(cpf == NULL)
   return;

 stepnum = 0;
 indexlimit = 0;
 for(i=0; i<STEP_PARTS; i++) {

   edgenum = 0;
   indexlimit += sp->indices_nr/STEP_PARTS;
   stepnum += sp->indices_nr/STEP_PARTS;

   while(cpf->index < indexlimit) {
    if(edgenum==0)
      print_condition(fp, sp->jspice_voltage_names[cpf->signum], cpf->time, 
         cpf->index, cpf->value, !runall, stepnum);
    else
      print_condition(fp, sp->jspice_voltage_names[cpf->signum], cpf->time, 
         cpf->index, cpf->value, 0, stepnum);
    edgenum ++;
    cpf=cpf->next;
    if(cpf==NULL)
     return;
   }

   if(edgenum>0)
     stepnum=0;
 } 
}


static int generate_async_latch_passfail(FILE*fp, SIGNALS *sp, int runall)
{
 TIME total_time, last_edge, next_edge, curr_time, check_time;
 int curr_index;
 int curr_value;
 int sig_num;
 EDGE *pp;
 int done;
 int last_edge_dir;
 TIME min_interval_to_edge;
 PF_CONDITION *bpf;

 total_time = sp->start_time+sp->indices_nr*sp->min_time_delta;

 for(sig_num=sp->signal_num-sp->outs_num; sig_num<sp->signal_num; sig_num++) 
  {
    curr_value = 0;
    last_edge = sp->start_time;
    pp = sp->sig_edge_bgn[sig_num];
    done = 0;

    do {
      if(pp != NULL) {
        last_edge_dir = 1-pp->direction;
        next_edge = pp->exact_time;
      }
      else {
        if(sp->sig_edge_bgn[sig_num]==NULL)
          last_edge_dir = 1 - sp->sig_initial[sig_num];
        else
          last_edge_dir = 1 - last_edge_dir;
        next_edge = total_time;
        done = 1;
      }

      if(last_edge_dir==RISING_EDGE)
        curr_time = last_edge + param.min_interval_to_riseedge;
      else
      if(last_edge_dir==FALLING_EDGE)
        curr_time = last_edge + param.min_interval_to_falledge;

      if(next_edge - last_edge < param.min_interval_to_riseedge + param.min_checks_interval + param.min_interval_to_falledge)
        curr_time = (next_edge + last_edge)/2;
       
      if(last_edge_dir==RISING_EDGE)
        min_interval_to_edge = param.min_interval_to_falledge;
      else
      if(last_edge_dir==FALLING_EDGE)
        min_interval_to_edge = param.min_interval_to_riseedge;

      while(curr_time < next_edge - min_interval_to_edge) {
        curr_index = (curr_time-sp->start_time)/sp->min_time_delta;
        check_time = sp->start_time + curr_index*sp->min_time_delta;

        collect_passfail_condition(&bpf, sig_num, check_time, curr_index, curr_value);

        curr_time += param.min_checks_interval;
      }
      curr_value = 1 - curr_value;
      last_edge = next_edge;
      if(pp!=NULL)
        pp=pp->next;

   } while (!done);

  }

 print_passfail_conditions(fp, sp, bpf, runall);
}


int generate_latch_passfail_file(SIGNALS *sp, char *filename, int runall)
/*
 runall = 1 - the whole simulation is run first, then the checking
              is performed;
          0 - simulation is run in smaller periods;
*/
{
 FILE *fp;

 fp=fopen(filename, "w");
 if(fp==NULL)
   return 0;

 fprintf(fp, "%s\n", filename);
 fprintf(fp, ".control\n\n");

 fprintf(fp, "vthreshold0 = %f\n", param.max_zero_volt);
 fprintf(fp, "vthreshold1 = %f\n\n", param.min_one_volt); 
 
 if(runall)
   fprintf(fp, "run\n\n");

 if(sp->clk_num>0)
   generate_sync_latch_passfail(fp, sp, runall);
 else
   generate_async_latch_passfail(fp, sp, runall);

 fprintf(fp, "if chFAIL = 1\n");
 fprintf(fp, "\tfree all yes\n");
 fprintf(fp, "\tchFAIL = 1\n");
 fprintf(fp, "end\n\n");

 fprintf(fp, "if chFAIL = 0\n");
 fprintf(fp, "\tfree all yes\n");
 fprintf(fp, "\tchFAIL = 0\n");
 fprintf(fp, "end\n\n");

 return 1;
}

