#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define DEBUG  0



#define FUNC(x) ((*func)(x))

static double trapzd(double (*func)(double), double a, double b, int n)
{
	double x,tnm,sum,del;
	static double s;
	static int it;
	int j;

	if (n == 1) {
		it=1;
		return (s=0.5*(b-a)*(FUNC(a)+FUNC(b)));
	} else {
		tnm=it;
		del=(b-a)/tnm;
		x=a+0.5*del;
		for (sum=0.0,j=1;j<=it;j++,x+=del) sum += FUNC(x);
		it *= 2;
		s=0.5*(s+(b-a)*sum/tnm);
		return s;
	}
}


#define EPS 1.0e-6
#define JMAX 20

static double qsimp(double (*func)(double), double a, double b)
{
	int j;
	double s,st,ost,os;

	ost = os =  -1.0e30;
	for (j=1;j<=JMAX;j++) {
		st=trapzd(func,a,b,j);
		s=(4.0*st-ost)/3.0;
		if (fabs(s-os) < EPS*fabs(os)) return s;
		os=s;
		ost=st;
	}
	nrerror("Too many steps in routine QSIMP");
}

#undef EPS
#undef JMAX


double normal(double x)
{
 double y;

 y=-x*x/2;
 y=exp(y)/sqrt(2*M_PI);
 return (double) y;
}


static dimensions;

static double multinormal(double x)
{
 int i;
 double mx, y;

 y=-x*x/2;
 mx=1.0;
 for(i=0;i<dimensions-1; i++)
   mx*=x;
 y=mx*exp(y);

 return y;
}


static int factorial(int n)
{
 if(n<=1)
   return 1;
 else
   return n*factorial(n-1);
}

static int double_factorial(int n)
{
 if(n<=1)
   return 1;
 else
   return n*factorial(n-2);
}

static long int power2(int n)
{
 long int ny=1;

 while(n--)
  ny*=2;

 return ny;
}

static double multicoeff(int n)
{
 int s, odd;

 s=(n-1)/2;
 odd=(n-1)%2;

 if(odd)
   return 1.0/(power2(s)*factorial(s));
 else
   return sqrt(2/M_PI)/double_factorial(2*s-1);
}


double area(double leftmargin, double rightmargin)
{
 return qsimp(normal, 0, leftmargin) + qsimp(normal, 0, rightmargin);
}


double multiarea(double r, int dim)
{
 double coeff, integral;

 dimensions = dim;
 integral=qsimp(multinormal, 0, r);
 coeff = multicoeff(dim);

 return integral*coeff;
}




#define IM	714025
#define IA	  4096
#define IC	 54773

#define SHUFFLE_LENGTH 98


double uniform_deviate(long *idum)
/*
  Return a uniform random deviate between 0.0 and 1.0.
  Set idum to any negative value to initialize or reinitialize the 
  sequance.
*/
{

 static long iy, shuff_tab[SHUFFLE_LENGTH];
 static int iff=0;

 int j;

 if((*idum < 0) || (iff==0)) {
	iff = 1;
	if((*idum = (IC-(*idum)) % IM) < 0)
		*idum = - (*idum);

	/* initialize the shuffle table */
	for(j=1; j<=SHUFFLE_LENGTH-1; j++) {
		*idum = (IA*(unsigned long)(*idum)+IC) % IM;
		shuff_tab[j] = (*idum);
 	}
	*idum = (IA*(unsigned long)(*idum)+IC) % IM;
	iy = (*idum);
 }
 j = 1 + (SHUFFLE_LENGTH-1)*iy/IM;

 if((j>SHUFFLE_LENGTH-1) || (j<0))
  nrerror("Error in random number generator uniform_deviate.");

 iy = shuff_tab[j];

 *idum = (IA*(unsigned long)(*idum)+IC) % IM;
 shuff_tab[j] = *idum;

 return (double) iy/ IM;
}


static double gauss_deviate(long *idum)
/*
  rightReturns a normally distributed deviate with zero mean and unit variance,
  using uniform_deviate() as the source of uniform deviates.
*/
{
 static int iset=0;
 static double gset;

 double fac, r, v1, v2;

 if(iset == 0) {
  	do {
  	  /* pick two uniform numbers in the square extending 
		from -1.0 to +1.0 in each direction */
  	    v1 = 2.0*uniform_deviate(idum) - 1.0;
 	    v2 = 2.0*uniform_deviate(idum) - 1.0;
 	    r = v1*v1 + v2*v2;
 	 } while(r>=1.0);

	/* Now make the Box-Muller transformation to get two
	   normal deviates. Return one and save the other for 
	   next time */
	fac = sqrt(-2.0*log(r)/r);

	gset = v1*fac;
	iset = 1;
	return v2*fac;
 }
 else {
	iset = 0;
	return gset;
 }

}

#if DEBUG
main()
{
double lm, rm;

printf("left margin=");
scanf("%lf", &lm);
printf("right margin=");
scanf("%lf", &rm);
printf("... %6.4f %6.4f\n",lm,rm);
printf("the answer is ... %6.4f\n",area(lm,rm));
}
#endif
